@@ -64,12 +64,13 @@ typedef edOptions *edOptionsPtr;
 
 typedef enum _XmlEdOp {
    XML_ED_DELETE,
+   XML_ED_VAR,
    XML_ED_INSERT,
    XML_ED_APPEND,
    XML_ED_UPDATE,
    XML_ED_RENAME,
    XML_ED_MOVE,
-   XML_ED_SUBNODE   
+   XML_ED_SUBNODE
 } XmlEdOp;
 
 /* TODO ??? */
@@ -83,7 +84,34 @@ typedef enum _XmlNodeType {
    XML_EXPR
 } XmlNodeType;
 
-typedef char* XmlEdArg;
+typedef struct {
+    char shortOpt;
+    const char* longOpt;        /* include "--" */
+    XmlNodeType type;
+} OptionSpec;
+
+static const OptionSpec
+    OPT_VAL_OR_EXP[] = {
+        {'x', "--expr", XML_EXPR},
+        {'v', "--value", XML_TEXT}
+    },
+    OPT_JUST_VAL[] = {
+        {'v', "--value", XML_TEXT}
+    },
+    OPT_JUST_TYPE[] = {
+        {'t', "--type"}
+    },
+    OPT_NODE_TYPE[] = {
+        {0, "elem", XML_ELEM},
+        {0, "attr", XML_ATTR},
+        {0, "text", XML_TEXT}
+    },
+    OPT_JUST_NAME[] = {
+        {'n', "--name"}
+    };
+
+
+typedef const char* XmlEdArg;
 
 typedef struct _XmlEdAction {
   XmlEdOp       op;
@@ -93,60 +121,16 @@ typedef struct _XmlEdAction {
   XmlNodeType   type;
 } XmlEdAction;
 
-#define MAX_XML_ED_OPS 128
-
-int ops_count = 0;
-static XmlEdAction ops[MAX_XML_ED_OPS];
-
-/*
- * usage string chunk : 509 char max on ISO C90
- */
-static const char edit_usage_str_1[] =
-"XMLStarlet Toolkit: Edit XML document(s)\n"
-"Usage: %s ed <global-options> {<action>} [ <xml-file-or-uri> ... ]\n"
-"where\n"
-"  <global-options>  - global options for editing\n"
-"  <xml-file-or-uri> - input XML document file name/uri (stdin otherwise)\n\n";
-
-
-static const char edit_usage_str_2[] =
-"<global-options> are:\n"
-"  -P, or -S           - preserve whitespace nodes.\n"
-"     (or --pf, --ps)    Note that space between attributes is not preserved\n"
-"  -O (or --omit-decl) - omit XML declaration (<?xml ...?>)\n"
-"  -L (or --inplace)   - edit file inplace\n"
-"  -N <name>=<value>   - predefine namespaces (name without \'xmlns:\')\n"
-"                        ex: xsql=urn:oracle-xsql\n"
-"                        Multiple -N options are allowed.\n"
-"                        -N options must be last global options.\n";
-
-static const char edit_usage_str_3[] =
-"  --net               - allow network access\n"
-"  --help or -h        - display help\n\n"
-"where <action>\n"
-"  -d or --delete <xpath>\n"
-"  -i or --insert <xpath> -t (--type) elem|text|attr -n <name> -v (--value) <value>\n"
-"  -a or --append <xpath> -t (--type) elem|text|attr -n <name> -v (--value) <value>\n"
-"  -s or --subnode <xpath> -t (--type) elem|text|attr -n <name> -v (--value) <value>\n";
-
-static const char edit_usage_str_4[] =
-"  -m or --move <xpath1> <xpath2>\n"
-"  -r or --rename <xpath1> -v <new-name>\n"
-"  -u or --update <xpath> -v (--value) <value>\n"
-"                         -x (--expr) <xpath>\n\n";
-
 /**
  *  display short help message
  */
-void
+static void
 edUsage(const char *argv0, exit_status status)
 {
+    extern void fprint_edit_usage(FILE* o, const char* argv0);
     extern const char more_info[];
     FILE *o = (status == EXIT_SUCCESS)? stdout : stderr;
-    fprintf(o, edit_usage_str_1, argv0);
-    fprintf(o, "%s", edit_usage_str_2);
-    fprintf(o, "%s", edit_usage_str_3);
-    fprintf(o, "%s", edit_usage_str_4);
+    fprint_edit_usage(o, argv0);
     fprintf(o, "%s", more_info);
     exit(status);
 }
@@ -154,7 +138,7 @@ edUsage(const char *argv0, exit_status status)
 /**
  *  Initialize global command line options
  */
-void
+static void
 edInitOptions(edOptionsPtr ops)
 {
     ops->noblanks = 1;
@@ -167,7 +151,7 @@ edInitOptions(edOptionsPtr ops)
 /**
  *  Parse global command line options
  */
-int
+static int
 edParseOptions(edOptionsPtr ops, int argc, char **argv)
 {
     int i;
@@ -213,7 +197,7 @@ edParseOptions(edOptionsPtr ops, int argc, char **argv)
 /**
  *  register the namespace from @ns_arr to @ctxt
  */
-void
+static void
 nsarr_xpath_register(xmlXPathContextPtr ctxt)
 {
     int ns;
@@ -238,16 +222,24 @@ extract_ns_defs(xmlDocPtr doc, xmlXPathContextPtr ctxt)
     }
 }
 
+static void
+update_string(xmlDocPtr doc, xmlNodePtr dest, const xmlChar* newstr)
+{
+    /* TODO: do we need xmlEncodeEntitiesReentrant() too/instead? */
+    xmlChar* string = xmlEncodeSpecialChars(doc, newstr);
+    xmlNodeSetContent(dest, string);
+    xmlFree(string);
+}
+
 /**
  *  'update' operation
  */
-void
+static void
 edUpdate(xmlDocPtr doc, xmlNodeSetPtr nodes, const char *val,
     XmlNodeType type, xmlXPathContextPtr ctxt)
 {
     int i;
     xmlXPathCompExprPtr xpath = NULL;
-    xmlNodePtr ctxt_node = ctxt->node;
 
     if (type == XML_EXPR) {
         xpath = xmlXPathCompile((const xmlChar*) val);
@@ -259,45 +251,98 @@ edUpdate(xmlDocPtr doc, xmlNodeSetPtr nodes, const char *val,
         /* update node */
         if (type == XML_EXPR) {
             xmlXPathObjectPtr res;
-            xmlChar *string;
 
             ctxt->node = nodes->nodeTab[i];
-            res = xmlXPathConvertString(xmlXPathCompiledEval(xpath, ctxt));
-            /* TODO: do we need xmlEncodeEntitiesReentrant() too/instead? */
-            string = xmlEncodeSpecialChars(doc, res->stringval);
+            res = xmlXPathCompiledEval(xpath, ctxt);
+            if (res->type == XPATH_NODESET || res->type == XPATH_XSLT_TREE) {
+                int j;
+                xmlNodePtr oldChild;
+                xmlNodeSetPtr oldChildren = xmlXPathNodeSetCreate(NULL);
+                /* NOTE: newChildren can be NULL for empty result set */
+                xmlNodeSetPtr newChildren = res->nodesetval;
+
+                /* NOTE: nodes can be both oldChildren and newChildren */
+
+                /* unlink the old children */
+                for (oldChild = nodes->nodeTab[i]->children; oldChild; oldChild = oldChild->next) {
+                    xmlUnlinkNode(oldChild);
+                    /* we can't free it now because an oldChild can also be
+                       newChild! just put it in the list */
+                    xmlXPathNodeSetAdd(oldChildren, oldChild);
+                }
+
+                /* add the new children */
+                for (j = 0; newChildren && j < newChildren->nodeNr; j++) {
+                    xmlNodePtr node = newChildren->nodeTab[j];
+                    xmlAddChild(nodes->nodeTab[i],
+                        /* if node is linked to this doc we need to copy */
+                        (node->doc == doc)? xmlDocCopyNode(node, doc, 1) : node);
+                    newChildren->nodeTab[j] = NULL;
+                }
+                newChildren->nodeNr = 0;
+
+                /* NOTE: if any oldChildren were newChildren, they've been
+                   copied so we can free them all now */
+                for (j = 0; j < oldChildren->nodeNr; j++) {
+                    xmlFreeNode(oldChildren->nodeTab[j]);
+                    oldChildren->nodeTab[j] = NULL;
+                }
+                oldChildren->nodeNr = 0;
+                xmlXPathFreeNodeSet(oldChildren);
+            } else {
+                res = xmlXPathConvertString(res);
+                update_string(doc, nodes->nodeTab[i], res->stringval);
+            }
             xmlXPathFreeObject(res);
-            xmlNodeSetContent(nodes->nodeTab[i], string);
-            xmlFree(string);
         } else {
-            /* TODO: do we need xmlEncodeEntitiesReentrant() too/instead? */
-            xmlChar *content = xmlEncodeSpecialChars(NULL, (const xmlChar*) val);
-            xmlNodeSetContent(nodes->nodeTab[i], content);
-            xmlFree(content);
+            update_string(doc, nodes->nodeTab[i], (const xmlChar*) val);
         }
     }
 
     xmlXPathFreeCompExpr(xpath);
-    ctxt->node = ctxt_node;
+}
+
+/* holds the node that was last inserted */
+static xmlNodeSetPtr previous_insertion;
+
+/**
+ * We must not keep free'd nodes in @previous_insertion.
+ * This is a callback from xmlFreeNode()
+ */
+static void
+removeNodeFromPrev(xmlNodePtr node)
+{
+    xmlXPathNodeSetDel(previous_insertion, node);
 }
 
 /**
  *  'insert' operation
  */
-void
+static void
 edInsert(xmlDocPtr doc, xmlNodeSetPtr nodes, const char *val, const char *name,
          XmlNodeType type, int mode)
 {
     int i;
+
+    xmlXPathEmptyNodeSet(previous_insertion);
+
     for (i = 0; i < nodes->nodeNr; i++)
     {
+        xmlNodePtr node;
+
+        if (nodes->nodeTab[i] == (void*) doc && mode != 0) {
+            fprintf(stderr, "The document node cannot have siblings.\n");
+            exit(EXIT_INTERNAL_ERROR);
+        }
+
         /* update node */
         if (type == XML_ATTR)
         {
-            xmlNewProp(nodes->nodeTab[i], BAD_CAST name, BAD_CAST val);
+            node = (xmlNodePtr) xmlNewProp(nodes->nodeTab[i], BAD_CAST name, BAD_CAST val);
         }
         else if (type == XML_ELEM)
         {
-            xmlNodePtr node = xmlNewDocNode(doc, NULL /* TODO: NS */, BAD_CAST name, BAD_CAST val);
+            node = xmlNewDocNode(doc, NULL /* TODO: NS */, BAD_CAST name, BAD_CAST val);
             if (mode > 0)
                 xmlAddNextSibling(nodes->nodeTab[i], node);
             else if (mode < 0)
@@ -307,7 +352,7 @@ edInsert(xmlDocPtr doc, xmlNodeSetPtr nodes, const char *val, const char *name,
         }
         else if (type == XML_TEXT)
         {
-            xmlNodePtr node = xmlNewDocText(doc, BAD_CAST val);
+            node = xmlNewDocText(doc, BAD_CAST val);
             if (mode > 0)
                 xmlAddNextSibling(nodes->nodeTab[i], node);
             else if (mode < 0)
@@ -315,18 +360,23 @@ edInsert(xmlDocPtr doc, xmlNodeSetPtr nodes, const char *val, const char *name,
             else
                 xmlAddChild(nodes->nodeTab[i], node);
         }
+        xmlXPathNodeSetAdd(previous_insertion, node);
     }
 }
 
 /**
  *  'rename' operation
  */
-void
-edRename(xmlDocPtr doc, xmlNodeSetPtr nodes, char *val, XmlNodeType type)
+static void
+edRename(xmlDocPtr doc, xmlNodeSetPtr nodes, const char *val, XmlNodeType type)
 {
     int i;
     for (i = 0; i < nodes->nodeNr; i++)
     {
+        if (nodes->nodeTab[i] == (void*) doc) {
+            fprintf(stderr, "The document node cannot be renamed.\n");
+            exit(EXIT_INTERNAL_ERROR);
+        }
         xmlNodeSetName(nodes->nodeTab[i], BAD_CAST val);
     }
 }
@@ -334,12 +384,17 @@ edRename(xmlDocPtr doc, xmlNodeSetPtr nodes, char *val, XmlNodeType type)
 /**
  *  'delete' operation
  */
-void
+static void
 edDelete(xmlDocPtr doc, xmlNodeSetPtr nodes)
 {
     int i;
     for (i = nodes->nodeNr - 1; i >= 0; i--)
     {
+        if (nodes->nodeTab[i] == (void*) doc) {
+            fprintf(stderr, "The document node cannot be deleted.\n");
+            exit(EXIT_INTERNAL_ERROR);
+        }
+
         if (nodes->nodeTab[i]->type == XML_NAMESPACE_DECL) {
             fprintf(stderr, "FIXME: can't delete namespace nodes\n");
             exit(EXIT_INTERNAL_ERROR);
@@ -356,12 +411,17 @@ edDelete(xmlDocPtr doc, xmlNodeSetPtr nodes)
 /**
  *  'move' operation
  */
-void
+static void
 edMove(xmlDocPtr doc, xmlNodeSetPtr nodes, xmlNodePtr to)
 {
     int i;
     for (i = 0; i < nodes->nodeNr; i++)
     {
+        if (nodes->nodeTab[i] == (void*) doc) {
+            fprintf(stderr, "The document node cannot be moved.\n");
+            exit(EXIT_INTERNAL_ERROR);
+        }
+
         if (nodes->nodeTab[i]->type == XML_NAMESPACE_DECL) {
             fprintf(stderr, "FIXME: can't move namespace nodes\n");
             exit(EXIT_INTERNAL_ERROR);
@@ -375,12 +435,20 @@ edMove(xmlDocPtr doc, xmlNodeSetPtr nodes, xmlNodePtr to)
 /**
  *  Loop through array of operations and perform them
  */
-void
-edProcess(xmlDocPtr doc, XmlEdAction* ops, int ops_count)
+static void
+edProcess(xmlDocPtr doc, const XmlEdAction* ops, int ops_count)
 {
     int k;
     xmlXPathContextPtr ctxt = xmlXPathNewContext(doc);
     /* NOTE: later registrations override earlier ones */
+    registerXstarNs(ctxt);
+
+    /* variables */
+    previous_insertion = xmlXPathNodeSetCreate(NULL);
+    registerXstarVariable(ctxt, "prev",
+        xmlXPathWrapNodeSet(previous_insertion));
+    xmlDeregisterNodeDefault(&removeNodeFromPrev);
+
 #if HAVE_EXSLT_XPATH_REGISTER
     /* register extension functions */
     exsltDateXpathCtxtRegister(ctxt, BAD_CAST "date");
@@ -392,13 +460,22 @@ edProcess(xmlDocPtr doc, XmlEdAction* ops, int ops_count)
     extract_ns_defs(doc, ctxt);
     /* namespaces from command line */
     nsarr_xpath_register(ctxt);
-    ctxt->node = xmlDocGetRootElement(doc);
 
     for (k = 0; k < ops_count; k++)
     {
         xmlXPathObjectPtr res;
         xmlNodeSetPtr nodes;
 
+        /* NOTE: to make relative paths match as if from "/", set context to
+           document; setting to root would match as if from "/node()/" */
+        ctxt->node = (xmlNodePtr) doc;
+
+        if (ops[k].op == XML_ED_VAR) {
+            res = xmlXPathEvalExpression(BAD_CAST ops[k].arg2, ctxt);
+            xmlXPathRegisterVariable(ctxt, BAD_CAST ops[k].arg1, res);
+            continue;
+        }
+
         res = xmlXPathEvalExpression(BAD_CAST ops[k].arg1, ctxt);
         if (!res || res->type != XPATH_NODESET || !res->nodesetval) continue;
         nodes = res->nodesetval;
@@ -409,8 +486,9 @@ edProcess(xmlDocPtr doc, XmlEdAction* ops, int ops_count)
                 edDelete(doc, nodes);
                 break;
             case XML_ED_MOVE: {
-                xmlXPathObjectPtr res_to =
-                    xmlXPathEvalExpression(BAD_CAST ops[k].arg2, ctxt);
+                xmlXPathObjectPtr res_to;
+                ctxt->node = (xmlNodePtr) doc;
+                res_to = xmlXPathEvalExpression(BAD_CAST ops[k].arg2, ctxt);
                 if (!res_to
                     || res_to->type != XPATH_NODESET
                     || res_to->nodesetval->nodeNr != 1) {
@@ -441,24 +519,29 @@ edProcess(xmlDocPtr doc, XmlEdAction* ops, int ops_count)
         }
         xmlXPathFreeObject(res);
     }
+    /* NOTE: free()ing ctxt also free()s previous_insertion */
+    previous_insertion = NULL;
+    xmlDeregisterNodeDefault(NULL);
+
     xmlXPathFreeContext(ctxt);
 }
 
 /**
  *  Output document
  */
-void
-edOutput(const char* filename, edOptions g_ops)
+static void
+edOutput(const char* filename, const XmlEdAction* ops, int ops_count,
+    const edOptions* g_ops)
 {
     xmlDocPtr doc;
     int save_options =
 #if LIBXML_VERSION >= 20708
-        (g_ops.noblanks? 0 : XML_SAVE_WSNONSIG) |
+        (g_ops->noblanks? 0 : XML_SAVE_WSNONSIG) |
 #endif
-        (g_ops.preserveFormat? 0 : XML_SAVE_FORMAT) |
-        (g_ops.omit_decl? XML_SAVE_NO_DECL : 0);
+        (g_ops->preserveFormat? 0 : XML_SAVE_FORMAT) |
+        (g_ops->omit_decl? XML_SAVE_NO_DECL : 0);
     int read_options =
-        (g_ops.nonet? XML_PARSE_NONET : 0);
+        (g_ops->nonet? XML_PARSE_NONET : 0);
     xmlSaveCtxtPtr save;
 
     doc = xmlReadFile(filename, NULL, read_options);
@@ -484,20 +567,74 @@ edOutput(const char* filename, edOptions g_ops)
         set_stdout_binary();
     }
 
-    save = xmlSaveToFilename(g_ops.inplace? filename : "-", NULL, save_options);
+    save = xmlSaveToFilename(g_ops->inplace? filename : "-", NULL, save_options);
     xmlSaveDoc(save, doc);
     xmlSaveClose(save);
 
     xmlFreeDoc(doc);
 }
 
+/**
+ * get next command line arg, or print error exit and exit if there isn't one
+ * @returns pointer to the arg
+ * @argi is incremented
+ */
+static const char*
+nextArg(char *const*const argv, int *argi)
+{
+    const char *arg = argv[*argi];
+    if (arg == NULL)
+    {
+        edUsage(argv[0], EXIT_BAD_ARGS);
+    }
+    *argi += 1;
+    return arg;
+}
+
+/**
+ * like nextArg(), but additionally look for next arg in @choices
+ */
+static XmlNodeType
+parseNextArg(char *const*const argv, int *argi,
+    const OptionSpec choices[], int choices_count)
+{
+    const char* arg = nextArg(argv, argi);
+    int i;
+    for (i = 0; i < choices_count; i++) {
+        if ((arg[0] == '-' && arg[1] == choices[i].shortOpt) ||
+            (strcmp(arg, choices[i].longOpt) == 0))
+            return choices[i].type;
+    }
+    edUsage(argv[0], EXIT_BAD_ARGS);
+    return 0;                   /* never reach here */
+}
+#define parseNextArg(argv, argi, choices) \
+    parseNextArg(argv, argi, choices, COUNT_OF(choices))
+
+
+/** --insert, --append, and --subnode all take the same arguments */
+static void
+parseInsertionArgs(XmlEdOp op_type, XmlEdAction* op,
+    char *const*const argv, int *argi)
+{
+    op->op = op_type;
+    op->arg1 = nextArg(argv, argi);
+    parseNextArg(argv, argi, OPT_JUST_TYPE);
+    op->type = parseNextArg(argv, argi, OPT_NODE_TYPE);
+    parseNextArg(argv, argi, OPT_JUST_NAME);
+    op->arg3 = nextArg(argv, argi);
+    parseNextArg(argv, argi, OPT_JUST_VAL);
+    op->arg2 = nextArg(argv, argi);
+}
+
 /**
  *  This is the main function for 'edit' option
  */
 int
 edMain(int argc, char **argv)
 {
-    int i, j, n, start = 0;
+    int i, ops_count, max_ops_count = 8, n, start = 0;
+    XmlEdAction* ops = xmlMalloc(sizeof(XmlEdAction) * max_ops_count);
     static edOptions g_ops;
     int nCount = 0;
 
@@ -511,214 +648,93 @@ edMain(int argc, char **argv)
     /*
      *  Parse command line and fill array of operations
      */
-    j = 0;
+    ops_count = 0;
     i = start + nCount;
 
     while (i < argc)
     {
-        if (!strcmp(argv[i], "-d") || !strcmp(argv[i], "--delete"))
+        const char *arg = nextArg(argv, &i);
+        if (arg[0] == '-')
         {
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);
-            ops[j].op = XML_ED_DELETE;
-            ops[j].arg1 = argv[i];
-            ops[j].arg2 = 0;
-            ops[j].type = XML_UNDEFINED;
-            j++;            
-        }
-        else if (!strcmp(argv[i], "-m") || !strcmp(argv[i], "--move"))
-        {
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);
-            ops[j].op = XML_ED_MOVE;
-            ops[j].arg1 = argv[i];
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);
-            ops[j].arg2 = argv[i];
-            ops[j].type = XML_UNDEFINED;
-            j++;
-        }
-        else if (!strcmp(argv[i], "-u") || !strcmp(argv[i], "--update"))
-        {
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);
-            ops[j].type = XML_UNDEFINED;
-            ops[j].op = XML_ED_UPDATE;
-            ops[j].arg1 = argv[i];
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);
-            if (strcmp(argv[i], "-v") && strcmp(argv[i], "--value") &&
-                strcmp(argv[i], "-x") && strcmp(argv[i], "--expr"))
-                edUsage(argv[0], EXIT_BAD_ARGS);
+            if (ops_count >= max_ops_count)
+            {
+                max_ops_count *= 2;
+                ops = xmlRealloc(ops, sizeof(XmlEdAction) * max_ops_count);
+            }
+            ops[ops_count].type = XML_UNDEFINED;
 
-            if (!strcmp(argv[i], "-x") || !strcmp(argv[i], "--expr"))
-                ops[j].type = XML_EXPR;
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);
-            ops[j].arg2 = argv[i];
-
-            j++;
-        }
-        else if (!strcmp(argv[i], "-r") || !strcmp(argv[i], "--rename"))
-        {
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);
-            ops[j].type = XML_UNDEFINED;
-            ops[j].op = XML_ED_RENAME;
-            ops[j].arg1 = argv[i];
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);
-            if (strcmp(argv[i], "-v") && strcmp(argv[i], "--value"))
-                edUsage(argv[0], EXIT_BAD_ARGS);
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);
-            ops[j].arg2 = argv[i];
-            j++;
-        }
-        else if (!strcmp(argv[i], "-i") || !strcmp(argv[i], "--insert"))
-        {
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);
-            ops[j].type = XML_UNDEFINED;
-            ops[j].op = XML_ED_INSERT;
-            ops[j].arg1 = argv[i];
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            if (strcmp(argv[i], "-t") && strcmp(argv[i], "--type")) edUsage(argv[0], EXIT_BAD_ARGS);;
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            if (!strcmp(argv[i], "elem"))
+            if (!strcmp(arg, "-d") || !strcmp(arg, "--delete"))
             {
-                ops[j].type = XML_ELEM;
+                ops[ops_count].op = XML_ED_DELETE;
+                ops[ops_count].arg1 = nextArg(argv, &i);
+                ops[ops_count].arg2 = 0;
             }
-            else if (!strcmp(argv[i], "attr"))
+            else if (!strcmp(arg, "--var"))
             {
-                ops[j].type = XML_ATTR;
+                ops[ops_count].op = XML_ED_VAR;
+                ops[ops_count].arg1 = nextArg(argv, &i);
+                ops[ops_count].arg2 = nextArg(argv, &i);
             }
-            else if (!strcmp(argv[i], "text"))
+            else if (!strcmp(arg, "-m") || !strcmp(arg, "--move"))
             {
-                ops[j].type = XML_TEXT;
+                ops[ops_count].op = XML_ED_MOVE;
+                ops[ops_count].arg1 = nextArg(argv, &i);
+                ops[ops_count].arg2 = nextArg(argv, &i);
             }
-            else edUsage(argv[0], EXIT_BAD_ARGS);;
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            if (strcmp(argv[i], "-n") && strcmp(argv[i], "--name")) edUsage(argv[0], EXIT_BAD_ARGS);;
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            ops[j].arg3 = argv[i];
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            if (strcmp(argv[i], "-v") && strcmp(argv[i], "--value")) edUsage(argv[0], EXIT_BAD_ARGS);;
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            ops[j].arg2 = argv[i];
-            j++;
-        }
-        else if (!strcmp(argv[i], "-a") || !strcmp(argv[i], "--append"))
-        {
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            ops[j].type = XML_UNDEFINED;
-            ops[j].op = XML_ED_APPEND;
-            ops[j].arg1 = argv[i];
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            if (strcmp(argv[i], "-t") && strcmp(argv[i], "--type")) edUsage(argv[0], EXIT_BAD_ARGS);;
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            if (!strcmp(argv[i], "elem"))
+            else if (!strcmp(arg, "-u") || !strcmp(arg, "--update"))
             {
-                ops[j].type = XML_ELEM;
+                ops[ops_count].op = XML_ED_UPDATE;
+                ops[ops_count].arg1 = nextArg(argv, &i);
+                ops[ops_count].type = parseNextArg(argv, &i, OPT_VAL_OR_EXP);
+                ops[ops_count].arg2 = nextArg(argv, &i);
             }
-            else if (!strcmp(argv[i], "attr"))
+            else if (!strcmp(arg, "-r") || !strcmp(arg, "--rename"))
             {
-                ops[j].type = XML_ATTR;
+                ops[ops_count].op = XML_ED_RENAME;
+                ops[ops_count].arg1 = nextArg(argv, &i);
+                ops[ops_count].type = parseNextArg(argv, &i, OPT_JUST_VAL);
+                ops[ops_count].arg2 = nextArg(argv, &i);
             }
-            else if (!strcmp(argv[i], "text"))
+            else if (!strcmp(arg, "-i") || !strcmp(arg, "--insert"))
             {
-                ops[j].type = XML_TEXT;
+                parseInsertionArgs(XML_ED_INSERT, &ops[ops_count], argv, &i);
             }
-            else edUsage(argv[0], EXIT_BAD_ARGS);;
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            if (strcmp(argv[i], "-n") && strcmp(argv[i], "--name")) edUsage(argv[0], EXIT_BAD_ARGS);;
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            ops[j].arg3 = argv[i];
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            if (strcmp(argv[i], "-v") && strcmp(argv[i], "--value")) edUsage(argv[0], EXIT_BAD_ARGS);;
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            ops[j].arg2 = argv[i];
-            j++;
-        }
-        else if (!strcmp(argv[i], "-s") || !strcmp(argv[i], "--subnode"))
-        {
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            ops[j].type = XML_UNDEFINED;
-            ops[j].op = XML_ED_SUBNODE;
-            ops[j].arg1 = argv[i];
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            if (strcmp(argv[i], "-t") && strcmp(argv[i], "--type")) edUsage(argv[0], EXIT_BAD_ARGS);;
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            if (!strcmp(argv[i], "elem"))
+            else if (!strcmp(arg, "-a") || !strcmp(arg, "--append"))
             {
-                ops[j].type = XML_ELEM;
+                parseInsertionArgs(XML_ED_APPEND, &ops[ops_count], argv, &i);
             }
-            else if (!strcmp(argv[i], "attr"))
+            else if (!strcmp(arg, "-s") || !strcmp(arg, "--subnode"))
             {
-                ops[j].type = XML_ATTR;
+                parseInsertionArgs(XML_ED_SUBNODE, &ops[ops_count], argv, &i);
             }
-            else if (!strcmp(argv[i], "text"))
+            else
             {
-                ops[j].type = XML_TEXT;
+                fprintf(stderr, "Warning: unrecognized option '%s'\n", arg);
             }
-            else edUsage(argv[0], EXIT_BAD_ARGS);;
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            if (strcmp(argv[i], "-n") && strcmp(argv[i], "--name")) edUsage(argv[0], EXIT_BAD_ARGS);;
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            ops[j].arg3 = argv[i];
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            if (strcmp(argv[i], "-v") && strcmp(argv[i], "--value")) edUsage(argv[0], EXIT_BAD_ARGS);;
-            i++;
-            if (i >= argc) edUsage(argv[0], EXIT_BAD_ARGS);;
-            ops[j].arg2 = argv[i];
-            j++;
+            ops_count++;
         }
         else
         {
-            if (argv[i][0] != '-')
-            {
-                break;
-            }
+            i--;                /* it was a filename, we didn't use it */
+            break;
         }
-
-        i++;
     }
 
-    ops_count = j;
-
     xmlKeepBlanksDefault(0);
 
     if ((!g_ops.noblanks) || g_ops.preserveFormat) xmlKeepBlanksDefault(1);
 
     if (i >= argc)
     {
-        edOutput("-", g_ops);
+        edOutput("-", ops, ops_count, &g_ops);
     }
     
     for (n=i; n<argc; n++)
     {
-        edOutput(argv[n], g_ops);
+        edOutput(argv[n], ops, ops_count, &g_ops);
     }
 
+    xmlFree(ops);
     cleanupNSArr(ns_arr);
     xmlCleanupParser();
     xmlCleanupGlobals();
