@@ -111,66 +111,6 @@ typedef struct {
     int nest;
 } template_option;
 
-
-/*
- * usage string chunk : 509 char max on ISO C90
- */
-static const char select_usage_str_1[] =
-"XMLStarlet Toolkit: Select from XML document(s)\n"
-"Usage: %s sel <global-options> {<template>} [ <xml-file> ... ]\n"
-"where\n"
-"  <global-options> - global options for selecting\n"
-"  <xml-file> - input XML document file name/uri (stdin is used if missing)\n"
-"  <template> - template for querying XML document with following syntax:\n\n";
-
-static const char select_usage_str_2[] =
-"<global-options> are:\n"
-"  -Q or --quiet             - do not write anything to standard output.\n"
-"  -C or --comp              - display generated XSLT\n"
-"  -R or --root              - print root element <xsl-select>\n"
-"  -T or --text              - output is text (default is XML)\n"
-"  -I or --indent            - indent output\n"
-"  -D or --xml-decl          - do not omit xml declaration line\n"
-"  -B or --noblanks          - remove insignificant spaces from XML tree\n";
-
-static const char select_usage_str_3[] =
-"  -E or --encode <encoding> - output in the given encoding (utf-8, unicode...)\n"
-"  -N <name>=<value>         - predefine namespaces (name without \'xmlns:\')\n"
-"                              ex: xsql=urn:oracle-xsql\n"
-"                              Multiple -N options are allowed.\n"
-"  --net                     - allow fetch DTDs or entities over network\n"
-"  --help                    - display help\n\n";
-
-static const char select_usage_str_4[] =
-"Syntax for templates: -t|--template <options>\n"
-"where <options>\n"
-"  -c or --copy-of <xpath>   - print copy of XPATH expression\n"
-"  -v or --value-of <xpath>  - print value of XPATH expression\n"
-"  -o or --output <string>   - output string literal\n"
-"  -n or --nl                - print new line\n"
-"  -f or --inp-name          - print input file name (or URL)\n"
-"  -m or --match <xpath>     - match XPATH expression\n"
-"  --var <name> <value> --break or\n"
-"  --var <name>=<value>      - declare a variable (referenced by $name)\n";
-
-static const char select_usage_str_5[] =
-"  -i or --if <test-xpath>   - check condition <xsl:if test=\"test-xpath\">\n"
-"  --elif <test-xpath>       - check condition if previous conditions failed\n"
-"  --else                    - check if previous conditions failed\n"
-"  -e or --elem <name>       - print out element <xsl:element name=\"name\">\n"
-"  -a or --attr <name>       - add attribute <xsl:attribute name=\"name\">\n"
-"  -b or --break             - break nesting\n"
-"  -s or --sort op xpath     - sort in order (used after -m) where\n";
-
-static const char select_usage_str_6[] =
-"  op is X:Y:Z, \n"
-"      X is A - for order=\"ascending\"\n"
-"      X is D - for order=\"descending\"\n"
-"      Y is N - for data-type=\"numeric\"\n"
-"      Y is T - for data-type=\"text\"\n"
-"      Z is U - for case-order=\"upper-first\"\n"
-"      Z is L - for case-order=\"lower-first\"\n\n";
-
 static const template_option
     OPT_TEMPLATE = { 't', "template" },
     OPT_COPY_OF  = { 'c', "copy-of", BAD_CAST "copy-of", {{BAD_CAST "select", TARG_XPATH}}, 0 },
@@ -211,58 +151,17 @@ void
 caseSortFunction(xsltTransformContextPtr ctxt, xmlNodePtr *sorts,
     int nbsorts);
 
-static const char select_usage_str_7[] =
-"There can be multiple --match, --copy-of, --value-of, etc options\n"
-"in a single template. The effect of applying command line templates\n"
-"can be illustrated with the following XSLT analogue\n\n"
-
-"xml sel -t -c \"xpath0\" -m \"xpath1\" -m \"xpath2\" -v \"xpath3\" \\\n"
-"        -t -m \"xpath4\" -c \"xpath5\"\n\n"
-
-"is equivalent to applying the following XSLT\n\n";
-
-static const char select_usage_str_8[] =
-"<?xml version=\"1.0\"?>\n"
-"<xsl:stylesheet version=\"1.0\" xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\">\n"
-"<xsl:template match=\"/\">\n"
-"  <xsl:call-template name=\"t1\"/>\n"
-"  <xsl:call-template name=\"t2\"/>\n"
-"</xsl:template>\n"
-"<xsl:template name=\"t1\">\n"
-"  <xsl:copy-of select=\"xpath0\"/>\n"
-"  <xsl:for-each select=\"xpath1\">\n"
-"    <xsl:for-each select=\"xpath2\">\n"
-"      <xsl:value-of select=\"xpath3\"/>\n"
-"    </xsl:for-each>\n"
-"  </xsl:for-each>\n"
-"</xsl:template>\n";
-
-static const char select_usage_str_9[] =
-"<xsl:template name=\"t2\">\n"
-"  <xsl:for-each select=\"xpath4\">\n"
-"    <xsl:copy-of select=\"xpath5\"/>\n"
-"  </xsl:for-each>\n"
-"</xsl:template>\n"
-"</xsl:stylesheet>\n\n";
-
 /**
  *  Print small help for command line options
  */
 void
 selUsage(const char *argv0, exit_status status)
 {
+    extern void fprint_select_usage(FILE* out, const char* argv0);
     extern const char more_info[];
     extern const char libxslt_more_info[];
     FILE *o = (status == EXIT_SUCCESS)? stdout : stderr;
-    fprintf(o, select_usage_str_1, argv0);
-    fprintf(o, "%s", select_usage_str_2);
-    fprintf(o, "%s", select_usage_str_3);
-    fprintf(o, "%s", select_usage_str_4);
-    fprintf(o, "%s", select_usage_str_5);
-    fprintf(o, "%s", select_usage_str_6);
-    fprintf(o, "%s", select_usage_str_7);
-    fprintf(o, "%s", select_usage_str_8);
-    fprintf(o, "%s", select_usage_str_9);
+    fprint_select_usage(o, argv0);
     fprintf(o, "%s", more_info);
     fprintf(o, "%s", libxslt_more_info);
     exit(status);
@@ -598,7 +497,7 @@ selPrepareXslt(xmlDocPtr style, selOptionsPtr ops, xmlChar *ns_arr[],
     root = xmlNewDocRawNode(style, NULL, BAD_CAST "stylesheet", NULL);
     xmlDocSetRootElement(style, root);
     xmlNewProp(root, BAD_CAST "version", BAD_CAST "1.0");
-    xslns = xmlNewNs(root, BAD_CAST "http://www.w3.org/1999/XSL/Transform", BAD_CAST "xsl");
+    xslns = xmlNewNs(root, XSLT_NAMESPACE, BAD_CAST "xsl");
     xmlSetNs(root, xslns);
 
     ns = 0;
@@ -701,16 +600,22 @@ selPrepareXslt(xmlDocPtr style, selOptionsPtr ops, xmlChar *ns_arr[],
 
     if (use_value_of) {
         xmlNodePtr value_of_template, for_each, value_of, param;
+        /* <xsl:template name='value-of-template' xmlns:xsl=XSLT_NS> */
         value_of_template = xmlNewChild(root, xslns, BAD_CAST "template", NULL);
         xmlNewProp(value_of_template, BAD_CAST "name", BAD_CAST "value-of-template");
+        /*   <xsl:param name='select'/> */
         param = xmlNewChild(value_of_template, xslns, BAD_CAST "param", NULL);
+        xmlNewProp(param, BAD_CAST "name", BAD_CAST "select");
+        /*   <xsl:value-of select='$select'/> */
         value_of = xmlNewChild(value_of_template, xslns, BAD_CAST "value-of", NULL);
         xmlNewProp(value_of, BAD_CAST "select", BAD_CAST "$select");
-        xmlNewProp(param, BAD_CAST "name", BAD_CAST "select");
+        /*   <xsl:for-each select='exslt:node-set($select)[position()&gt;1]'> */
         for_each = xmlNewChild(value_of_template, xslns, BAD_CAST "for-each", NULL);
         xmlNewProp(for_each, BAD_CAST "select", BAD_CAST "exslt:node-set($select)[position()>1]");
+        /*     <xsl:value-of select='\"&#10;\"'/> */
         value_of = xmlNewChild(for_each, xslns, BAD_CAST "value-of", NULL);
         xmlNewProp(value_of, BAD_CAST "select", BAD_CAST "'\n'");
+        /*     <xsl:value-of select='.'/> */
         value_of = xmlNewChild(for_each, xslns, BAD_CAST "value-of", NULL);
         xmlNewProp(value_of, BAD_CAST "select", BAD_CAST ".");
     }
